"""
Copyright © 2021-2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""

import asyncio
import time
from typing import Optional, List, Callable
from uuid import UUID

import pyDE1
from pyDE1.config import config
from pyDE1.de1 import DE1
from pyDE1.de1.c_api import API_MachineStates
from pyDE1.de1.events import StateUpdate
from pyDE1.dispatcher.resource import ConnectivityEnum
from pyDE1.event_manager.event_manager import SubscribedEvent
from pyDE1.event_manager.events import (
    DeviceAvailabilityState, DeviceRole,
)
from pyDE1.exceptions import DE1NoAddressError, DE1APIValueError
from pyDE1.scale.generic_scale import GenericScale

from pyDE1.scale.events import (
    ScaleWeightUpdate, ScaleTareSeen, WeightAndFlowUpdate
)
from pyDE1.scanner import (
    find_first_matching,
    RegisteredPrefixes,
)
from pyDE1.singleton import Singleton

logger = pyDE1.getLogger('Scale.Processor')


class ScaleProcessor (Singleton):
    """
    Subscribes to weight-update events from a scale
    Provides estimates of weight and mass-flow via
    WeightAndFlowUpdate events

    Should be able to init and "wire to" without a scale
    Scales should be able to be changed on the fly
    """

    # NB: This is intentionally done in _singleton_init() and not __init__()
    #     See Singleton and Guido's notes there
    #
    # def __init__(self):
    #     pass

    def _singleton_init(self):

        self._scale = GenericScale()
        self._scale_weight_update_id: Optional[UUID] = None
        self._scale_tare_seen_id: Optional[UUID] = None
        self._scale_changed_id: Optional[UUID] = None
        self._state_update_id: Optional[UUID] = None
        self._history_max = 10  # Will be extended if needed by Estimator
        self._history_time: List[float] = []
        self._history_weight: List[float] = []
        self._history_lock = asyncio.Lock()
        # set_scale needs _history_lock
        self._event_weight_and_flow_update = SubscribedEvent(self)

        self.CURRENT_WEIGHT_MAX_AGE = 1.0  # seconds, else return None

        # init of Estimator checks that the targets are present
        # (good practice to explicily declare anyways)
        self._current_weight: float = 0
        self._current_weight_time: float = 0
        self._average_flow: float = 0
        self._average_flow_time: float = 0
        self._median_weight: float = 0
        self._median_weight_time: float = 0
        self._median_flow: float = 0
        self._median_flow_time: float = 0
        self._estimators = [
            CurrentWeight(self, '_current_weight'),
            AverageFlow(self, '_average_flow', 11),
            MedianWeight(self, '_median_weight', 11),
            MedianFlow(self, '_median_flow', 11, 5),
        ]

        asyncio.get_running_loop().create_task(self.wire_scale())

    @property
    def scale(self):
        return self._scale

    async def wire_scale(self):
        (
            self._scale_weight_update_id,
            self._scale_tare_seen_id,
            self._scale_changed_id,
            self._state_update_id,

        ) = await asyncio.gather(
            self._scale.event_weight_update.subscribe(
                self._weight_update_subscriber),

            self._scale.event_tare_seen.subscribe(
                self._tare_seen_subscriber),

            self._scale.event_scale_changed.subscribe(
                self._reset()
            ),

            DE1().event_state_update.subscribe(
                self._state_update_subscriber),

            return_exceptions=True
        )
        await self._reset()  # New scale, toss old history

    # Provide "null-safe" methods for API

    @property
    def scale_address(self):
        if self.scale is not None:
            return self.scale.address
        else:
            return None

    @property
    def scale_name(self):
        if self.scale is not None:
            return self.scale.name
        else:
            return None

    @property
    def scale_connectivity(self):
        if self.scale is not None:
            return self.scale.connectivity
        else:
            return ConnectivityEnum.NOT_CONNECTED

    async def connectivity_setter(self, value):
        if self.scale is not None:
            return await self.scale.connectivity_setter(value)
        else:
            raise DE1NoAddressError("No scale associated")

    @property
    def scale_availability_state(self):
        if self.scale is not None:
            return self.scale.availability_state
        else:
            return DeviceAvailabilityState.UNKNOWN

    async def scale_availability_setter(self, value):
        if self.scale is not None:
            return await self.scale.availability_setter(value)
        else:
            raise DE1NoAddressError("No scale associated")

    @property
    def device_availability_last_sent(self):
        return self.scale.device_availability_last_sent

    @property
    def current_weight(self) -> Optional[float]:
        if (time.time() - self._current_weight_time
                < self.CURRENT_WEIGHT_MAX_AGE):
            return self._current_weight
        else:
            return None



    #
    # Self-contained call for API
    #

    async def connect_to_first_if_found(self):
        if self.scale.is_connected:
            logger.warning(
                "'scan' requested, but already connected. "
                "No action taken.")
        else:
            device = await find_first_matching(DeviceRole.SCALE)
            if device:
                await self.scale.change_address(device)
                await self.scale.capture()
        return self.scale_address


    async def change_scale_to_id(self, ble_device_id: Optional[str]):
        """
        For now, this won't return until connected or fails to connect
        As a result, will trigger the timeout on API calls
        """
        logger.info(f"Request to replace scale with {ble_device_id}")

        if ble_device_id == 'scan':
            await self.connect_to_first_if_found()
        else:
            await self.scale.change_address(ble_device_id)

        return self.scale.address

    @property
    def event_weight_and_flow_update(self):
        return self._event_weight_and_flow_update

    async def _reset(self):
        async with self._history_lock:
            self._reset_have_lock()

    def _reset_have_lock(self):
        self._history_time = []
        self._history_weight = []
        # TODO: Perhaps should clear any pending updates
        #       as they may be pre-tare

    @property
    def _history_available(self):
        # Ultra safe
        return min(len(self._history_weight), len(self._history_time))

    async def _tare_seen_subscriber(self, sts: ScaleTareSeen):
        await self._reset()

    async def _weight_update_subscriber(self, swu: ScaleWeightUpdate):
        # TODO: This has the potential to get really messy
        #       Make sure this doesn't block other things
        # The Skale can return multiple updates in milliseconds
        # Is there any guarantee that they will be processed in order?
        # "Acquiring a lock is fair: the coroutine that proceeds
        #  will be the first coroutine that started waiting on the lock."
        async with self._history_lock:
            # Detect a gap in reporting being "too long"
            # (typically from a disconnect/reconnect)
            # A skip of three at 150 ms per update with the Skale II
            TOO_LONG = 0.8 # seconds
            try:
                if ((dt := swu.scale_time
                           - self._history_time[-1]) > TOO_LONG):
                    logger.warning(
                        "Resetting scale due to gap in reports: "
                        f"{dt:0.3f} > {TOO_LONG} s")
                    self._reset_have_lock()
                    return
            except IndexError:
                pass  # (No elements in the history list)
            self._history_time.append(swu.scale_time)
            self._history_weight.append(swu.weight)
            # Unlikely, but possibly the case that history_max shrunk
            while len(self._history_time) \
                    > self._history_max:
                self._history_time.pop(0)
            while len(self._history_weight) \
                    > self._history_max:
                self._history_weight.pop(0)

            # There's nothing here really parallelizable
            for estimator in self._estimators:
                estimator.estimate()

            await self._event_weight_and_flow_update.publish(
                WeightAndFlowUpdate(
                    arrival_time=swu.arrival_time,
                    scale_time=swu.scale_time,
                    current_weight=self._current_weight,
                    current_weight_time=self._current_weight_time,
                    average_flow=self._average_flow,
                    average_flow_time=self._average_flow_time,
                    median_weight=self._median_weight,
                    median_weight_time=self._median_weight_time,
                    median_flow=self._median_flow,
                    median_flow_time=self._median_flow_time,
                )
            )

    async def _state_update_subscriber(self, su: StateUpdate):
        scale = self.scale
        if (su.previous_state == API_MachineStates.Sleep
                and su.state != API_MachineStates.Sleep
                and scale.address not in ('', None)
                and not scale.is_captured):
            logger.info(
                "Reconnecting to scale as DE1 left Sleep to "
                f"{API_MachineStates(su.state).name}")
            await self.scale.request_capture()
        # If DE1 reports Sleep while the scale is awaiting connection
        # such as in change_scale_to_id(), a timeout will occur.
        # This happens when clicking both "Scan" buttons.
        elif (su.state == API_MachineStates.Sleep
              and scale.is_captured):
            logger.info(
                "Releasing scale as DE1 entered Sleep,"
                f"{API_MachineStates(su.substate).name}")
            await self.scale.request_release()

# Prevent dreaded "circular import" problems
from pyDE1.scale.estimator import CurrentWeight, AverageFlow, \
    MedianWeight, MedianFlow



"""
Copyright © 2021-2022 Jeff Kletsky. All Rights Reserved.

Utilities that are specific to pyDE1 DE1 and scales

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""
import asyncio
import logging
import time
from typing import Optional, Union

from bleak import BLEDevice

import pyDE1.task_logger

from pyDE1.bledev.managed_bleak_client import CaptureQueue, CaptureRequest, \
    ManagedBleakClient, cq_to_code
from pyDE1.btcontrack import persist_connection_file, remove_connection_file
from pyDE1.dispatcher.resource import ConnectivityEnum
from pyDE1.event_manager.event_manager import SubscribedEvent
from pyDE1.event_manager.events import (
    ConnectivityState, ConnectivityChange,
    DeviceAvailabilityState, DeviceAvailability, DeviceRole,
)

# The behavior of sending a ConnectivityChange notice is the same
# between the DE1 and the scales.
# It is assumed that there is
from pyDE1.exceptions import DE1APIValueError, DE1ValueError
from pyDE1.lock_logger import LockLogger
from pyDE1.utils import call_str, EventReadOnly


class ManagedBleakDevice:
    """
    Mixin to provide the behavior associated with a ManagedBleakClient
    in the context of DE1 or Scale implementations
    """

    # Renames
    #   prepare_for_connection          _prepare_for_connection
    #   initialize_after_connection     _initialize_after_connection
    #   _bleak_client                   _client

    def __init__(self):
        if not hasattr(self, 'logger'):
            self.logger = logging.getLogger(self.__class__.__name__ )
        if not hasattr(self, '_role'):
            self._role = DeviceRole.UNKNOWN
        if not hasattr(self, '_name'):
            self._name = ''

        self._bleak_client: ManagedBleakClient = ManagedBleakClient(
            address_or_ble_device='',
            disconnected_callback=self._create_disconnected_callback(),
            on_change_callback=self._create_on_change_callback(),
            logger=self.logger.getChild('Client'),
        )

        self._event_connectivity = SubscribedEvent(
            self, adjust_payload=_resend_last_state_if_none)
        self._event_availability = SubscribedEvent(self)
        self._ready = asyncio.Event()
        self._ready_ro = EventReadOnly(self._ready)

        self._prepare_for_connection()

    def _prepare_for_connection(self):
        """
        Override in subclass as needed
        """
        pass

    async def _initialize_after_connection(self, hold_ready=False):
        """
        Override in subclass as needed
        :param hold_ready: Do not call _notify_ready() here if True
        """
        if not hold_ready:
            self._notify_ready()

    @property
    def is_captured(self) -> bool:
        return self._bleak_client.is_captured

    @property
    def is_released(self):
        return self._bleak_client.is_released

    @property
    def is_ready(self):
        return self._ready.is_set()

    @property
    def is_connected(self):
        """
        Legacy -- likely want is_captured or is_ready
        """
        return self._bleak_client.is_connected

    @property
    def connectivity_task_pending(self) -> bool:
        return self._bleak_client.connectivity_task_pending

    @property
    def active_request(self) -> Optional[CaptureRequest]:
        return self._bleak_client.active_request

    @property
    def event_captured(self) -> EventReadOnly:
        return self._bleak_client.event_captured

    @property
    def event_released(self) -> EventReadOnly:
        return self._bleak_client.event_released

    @property
    def event_ready(self) -> EventReadOnly:
        return self._ready_ro

    async def capture(self, timeout: Optional[float] = None) -> bool:
        return await self._bleak_client.capture(timeout=timeout)

    async def release(self, timeout: Optional[float] = None) -> bool:
        return await self._bleak_client.release(timeout=timeout)

    async def request_capture(self):
        return await self._bleak_client.request_capture()

    async def request_release(self):
        return await self._bleak_client.request_release()

    @property
    def connectivity_state(self)-> ConnectivityState:
        if self.is_ready:
            cs = ConnectivityState.READY
        else:
            cs = cq_to_cs(self._bleak_client._capture_queue)
        return cs

    @property
    def availability_state(self)-> DeviceAvailabilityState:
        das = cq_to_das(self._bleak_client._capture_queue)
        if das == DeviceAvailabilityState.CAPTURED and self.is_ready:
            das = DeviceAvailabilityState.READY
        return das

    @property
    def name(self):
        name = self._name if self._name else "{}_{}".format(
            self.__class__.__name__,
            self.address[-8:] if self.address else 'None')
        return name

    @property
    def address(self):
        return self._bleak_client.address

    @property
    def role(self):
        return self._role

    async def change_address(self,
                             address: Optional[Union[BLEDevice, str]]) -> bool:
        if isinstance(address, BLEDevice):
            self._name = address.name
        else:
            self._name = None
        changed = await self._bleak_client.change_address(address)
        if changed:
            self._event_connectivity.last_sent_clear()
        return changed

    ###
    ### End of public API
    ###

    # For HTTP API

    @property
    def connectivity(self):
        retval = ConnectivityEnum.NOT_CONNECTED
        if self.is_captured:
            if self._ready.is_set():
                retval = ConnectivityEnum.READY
            else:
                retval = ConnectivityEnum.CONNECTED
        return retval

    async def availability_setter(self, value):
        assert isinstance(value, CaptureRequest), \
            f"mode of {value} not a CaptureRequest "
        if value is CaptureRequest.CAPTURE:
            await self.request_capture()
        elif value is CaptureRequest.RELEASE:
            await self.request_release()
        else:
            raise DE1APIValueError(
                "Only CAPTURE and RELEASE can be set, not {value}")

    async def connectivity_setter(self, value):
        assert isinstance(value, ConnectivityEnum), \
            f"mode of {value} not a ConnectivityEnum "
        if value is ConnectivityEnum.CONNECTED:
            await self.request_capture()
        elif value is ConnectivityEnum.NOT_CONNECTED:
            await self.request_release()
        else:
            raise DE1APIValueError(
                "Only CONNECTED and NOT_CONNECTED can be set, "
                f"not {value}")

    # Internals

    def _create_disconnected_callback(self):
        """
        Override as needed
        """
        # device = self

        # def disconnected_callback(client: BleakClient) -> None:
        #     pass
        #
        # return disconnected_callback

        return None

    def _create_on_change_callback(self):

        def on_change_callback(client: ManagedBleakClient,
                               previous: CaptureQueue,
                               current: CaptureQueue) -> None:
            arrival_time = time.time()

            self.logger.debug(
                "[{:05.3f}] on_change_callback(<MBC>, {}, {}) {}".format(
                    arrival_time % 10,
                    cq_to_code(previous),
                    cq_to_code(current),
                    call_str()
                ))


            asyncio.create_task(self._on_change_callback_async(
                arrival_time=arrival_time,
                client=client,
                previous=previous,
                current=current))

        return on_change_callback

    async def _on_change_callback_async(self,
                                        arrival_time: Optional[float],
                                        client: ManagedBleakClient,
                                        previous: CaptureQueue,
                                        current: CaptureQueue) -> None:

        if arrival_time is None:
            arrival_time= time.time()

        previous_das = cq_to_das(previous)
        current_das = cq_to_das(current)

        previous_cs = cq_to_cs(previous)
        current_cs = cq_to_cs(current)


        # These are strings for logging, not used for logic
        pss = previous_cs if previous_cs not in (
            ConnectivityState.UNKNOWN, None) else f"({cq_to_code(previous)})"
        css = current_cs if current_cs not in (
            ConnectivityState.UNKNOWN, None) else f"({cq_to_code(current)})"

        self.logger.debug(f"Change from {pss} to {css}")

        # NB: Even though the CaptureQueue may have changed,
        # the ConnectivityState may not have changed

        if (self.is_ready
                and current_cs != ConnectivityState.CONNECTED):
            self._notify_not_ready()

        if (not self.is_ready
                and current_cs == ConnectivityState.CONNECTED
                and previous_cs != ConnectivityState.CONNECTED):
            pyDE1.task_logger.create_task(
                self._initialize_after_connection(),
                logger=self.logger,
                message=f"Exception in initialize_after_connection()")

        if (current_cs == ConnectivityState.DISCONNECTED
                and previous_cs != ConnectivityState.DISCONNECTED):
            self._prepare_for_connection()

        self._send_device_availability(arrival_time=arrival_time,
                                       new_state=current_das)
        self._send_connectivity_change(arrival_time=arrival_time,
                                       new_state=current_cs)

        if current_cs in (ConnectivityState.CONNECTING,
                          ConnectivityState.CONNECTED,
                          ConnectivityState.READY):
            try:
                persist_connection_file(self.address)
            except DE1ValueError as e:
                self.logger.exception(
                    "Connection filename error: "
                    f"{self.name} {self.role} at {self.address}",
                    exc_info=e,
                )
                raise e
        elif current_cs == ConnectivityState.DISCONNECTED:
            remove_connection_file(self.address)

    # Helper method to populate a ConnectivityChange

    def _connectivity_change(self, arrival_time: float,
                             state: ConnectivityState):
        return ConnectivityChange(arrival_time=arrival_time,
                                  state=state,
                                  id=self.address,
                                  name=self.name)

    def _send_connectivity_change(self, arrival_time: float,
                                  new_state: ConnectivityState):
        """
        Converts the current CaptureQueue into a ConnectivityState
        and creates a task to send it using _connectivity_change.
        If CaptureState.UNKNOWN returned by cq_to_cs(),
        use the previously-sent value if there is one.
        """

        send_state = new_state

        if send_state == ConnectivityState.UNKNOWN:
            try:
                send_state = self._event_connectivity.last_sent.state
            except AttributeError:
                pass
                # This is expected on the first call after initialization
                # self._logger.warning(
                #     "Connectivity last sent of "
                #     f"{self._event_connectivity.last_sent}, "
                #     "using UNKNOWN rather than last-sent state")

        asyncio.create_task(
            self._event_connectivity.publish(
                self._connectivity_change(arrival_time=arrival_time,
                                          state=send_state)))

    @property
    def device_availability_last_sent(self):
        return self._event_availability.last_sent.as_json()

    def _device_availability(self, arrival_time: float,
                             state: DeviceAvailabilityState):
        return DeviceAvailability(arrival_time=arrival_time,
                                  role=self.role,
                                  state=state,
                                  id=self.address,
                                  name=self.name)

    def _send_device_availability(self, arrival_time: float,
                                  new_state: DeviceAvailabilityState):
        """
        Sends DeviceAvailability
        """
        asyncio.create_task(
            self._event_availability.publish(
                self._device_availability(arrival_time=arrival_time,
                                          state=new_state)))


    def _notify_ready(self):
        self._ready.set()
        # Send the same way to prevent things from getting out of order
        self._send_device_availability(
            arrival_time=time.time(),
            new_state=DeviceAvailabilityState.READY)
        self._send_connectivity_change(arrival_time=time.time(),
                                       new_state=ConnectivityState.READY)

        self.logger.info("Ready")

    def _notify_not_ready(self):
        self._ready.clear()
        self._send_device_availability(
            arrival_time=time.time(),
            new_state=DeviceAvailabilityState.NOT_READY)
        self._send_connectivity_change(arrival_time=time.time(),
                                       new_state=ConnectivityState.NOT_READY)

class ClassChangeException(RuntimeError):
    pass

class ClassChangeLeaveException(ClassChangeException):
    pass

class ClassChangeAdoptException(ClassChangeException):
    pass


# Use as decorator, such as
#    @class_changer_generic_class
#    class ParentClass (ManagedBleakDevice, ClassChanger):
#        pass
def class_changer_generic_class(cls: 'ClassChanger'):
    cls._class_changer_generic_class = cls
    return cls

class ClassChanger:
    """
    NB: This should "be first" in the list of inherited classes
    See: https://stackoverflow.com/questions/9575409/
        calling-parent-class-init-with-multiple-inheritance-whats-the-right-way

    This is a mixin class that supports a limited ability to change between
    a generic parent class and various subclasses.  The intended application
    is for scales where the desired subclass is not able to be determined
    until the signature is seen over Bluetooth.  Changing out the object
    would either need all references to the prior instance to be updated
    or a proxy object hiding the instance of the moment behind an unchanging
    set of references for consumers.

    The basic flow of changing class is:
        * instance._leave_class() is called while the current class is known
        * instance.__class__ gets set to a new subclass
        * instance._adopt_class() is called to set up the instance,
          ready to call instance._initialize_after_connection()
          as the new class

    Calling _leave_class() on an instance should restore the instance to a state
    that is similar to that returned by GenericClass().  As this will sometimes
    be called as a result of establishing a Bluetooth connection, the state
    of that connection probably should not be modified.  This behavior is only
    available when a subclass of cls._class_changer_generic_class and not that
    class itself.

    Calling _adopt_class() on an instance should update either an instance
    of GenericClass or one on which _leave_class() has just been called
    to one that is fully functional based on its recently changed class
    and ready for _initialize_on_connect() to be called.

    Because of this class-changing behavior, some functionality isn't available
    until the class change completes. This is TBD right now.

    NB: This is fragile voodoo and subject to the internals of Python changing
    """

    # Set to the generic class for the hierarchy
    # Python 3.11 and later supports typing.Self

    _class_changer_generic_class = None

    def __init__(self, *args, **kwargs):
        super(ClassChanger, self).__init__(*args, **kwargs)
        self._class_change_lock = asyncio.Lock()

    def _is_generic_class(self):
        return type(self) == self._class_changer_generic_class

    def _is_descendant_class(self):
        return (not self._is_generic_class()
                and isinstance(self, self._class_changer_generic_class))

    async def _leave_class(self):
        if self._is_generic_class():
            return
        else:
            # do the work here
            pass

    async def _adopt_class(self):
        # should be able to adopt the generic class, even if a noop
        # do the work here
        pass

    async def _change_class(self, new_class: type):
        old_class_name = self.__class__.__name__
        new_class_name = new_class.__name__
        try:
            logger = self.logger
        except AttributeError:
            logger = logging.getLogger(old_class_name)
        if not issubclass(new_class, self._class_changer_generic_class):
            raise ClassChangeException(
                "Can't change from {} to {} as not a {}".format(
                    type(self), new_class, self._class_changer_generic_class
                ))
        ll = LockLogger(self._class_change_lock, 'ClassChange').check()
        async with self._class_change_lock:
            ll.acquired()
            logger.info(
                f"Leaving {old_class_name} for {new_class_name}")
            await self._leave_class()
            self.__class__ = new_class
            await self._adopt_class()
            logger.info(
                f"Changed from {old_class_name} to {new_class_name}")
        ll.released()


def cq_to_cs(cq: CaptureQueue) -> Optional[ConnectivityState]:
    """
    Given a CaptureQueue instance, return the "matching" ConnectivityState
    or, if indeterminate what will happen next (cancel pending), return None
    """
    cs = ConnectivityState.UNKNOWN

    # Nothing pending, current matches target

    if cq == CaptureQueue(CaptureRequest.CAPTURE, None,
                          CaptureRequest.CAPTURE):
        cs = ConnectivityState.CONNECTED

    elif cq == CaptureQueue(CaptureRequest.RELEASE, None,
                            CaptureRequest.RELEASE):
        cs = ConnectivityState.DISCONNECTED

    # Nothing pending, current != target

    elif cq.pending is None:

        if cq.target == CaptureRequest.CAPTURE:
            cs = ConnectivityState.CONNECTING

        elif cq.target == CaptureRequest.RELEASE:
            cs = ConnectivityState.DISCONNECTING

        # Nothing pending, no target

        elif cq.target is None:
            if cq.connected is None:
                cs = ConnectivityState.INITIAL  # Initial state
            elif cq.connected == CaptureRequest.CAPTURE:

                cs = ConnectivityState.CONNECTED
            elif cq.connected == CaptureRequest.RELEASE:
                cs = ConnectivityState.DISCONNECTED

    # Something pending

    elif cq.pending == CaptureRequest.CAPTURE:
        cs = ConnectivityState.CONNECTING

    elif cq.pending == CaptureRequest.RELEASE:
        cs = ConnectivityState.DISCONNECTING

    # Pending cancel
    # No information in cq as to what it was before, return None

    elif cq.pending == CaptureRequest.CANCEL:
        cs = ConnectivityState.UNKNOWN

    else:
        raise RuntimeError(f"Logic fall-through with {cq}")

    return cs


def cq_to_das(cq: CaptureQueue) -> DeviceAvailabilityState:
    """
    Given a CaptureQueue instance, return the "matching" DeviceAvailabilityState

    NB: This never returns READY or NOT_READY
    """
    das = DeviceAvailabilityState.UNKNOWN

    # Nothing pending, current matches target

    if cq == CaptureQueue(CaptureRequest.CAPTURE, None,
                          CaptureRequest.CAPTURE):
        das = DeviceAvailabilityState.CAPTURED

    elif cq == CaptureQueue(CaptureRequest.RELEASE, None,
                            CaptureRequest.RELEASE):
        das = DeviceAvailabilityState.RELEASED

    # Nothing pending, current != target

    elif cq.pending is None:

        if cq.target == CaptureRequest.CAPTURE:
            das = DeviceAvailabilityState.CAPTURING

        elif cq.target == CaptureRequest.RELEASE:
            das = DeviceAvailabilityState.RELEASING

        # Nothing pending, no target

        elif cq.target is None:
            if cq.connected is None:
                das = DeviceAvailabilityState.INITIAL  # Initial state
            elif cq.connected == CaptureRequest.CAPTURE:

                das = DeviceAvailabilityState.CAPTURED
            elif cq.connected == CaptureRequest.RELEASE:
                das = DeviceAvailabilityState.RELEASED

    # Something pending

    elif cq.pending == CaptureRequest.CAPTURE:
        das = DeviceAvailabilityState.CAPTURING

    elif cq.pending == CaptureRequest.RELEASE:
        das = DeviceAvailabilityState.RELEASING

    # Pending cancel

    elif cq.pending == CaptureRequest.CANCEL:
        if cq.target == CaptureRequest.CAPTURE:
            das = DeviceAvailabilityState.CAPTURING
        elif cq.target == CaptureRequest.RELEASE:
            das = DeviceAvailabilityState.RELEASING
        else:
            das = DeviceAvailabilityState.UNKNOWN

    else:
        raise RuntimeError(f"Logic fall-through with {cq}")

    return das


def _resend_last_state_if_none(se: SubscribedEvent,
                               payload: ConnectivityChange):
    if payload.state is ConnectivityState.UNKNOWN:
        if (old := se.last_sent) is not None:
            payload.state = old.state
        else:
            payload.state = ConnectivityState.UNKNOWN




"""
Copyright © 2021-2022 Jeff Kletsky. All Rights Reserved.

License for this software, part of the pyDE1 package, is granted under
GNU General Public License v3.0 only
SPDX-License-Identifier: GPL-3.0-only
"""
import asyncio
import enum
import struct
import time

from struct import unpack_from, pack
from typing import Optional, Union, NamedTuple

import pyDE1  # provides getLogger() and that's all
import pyDE1.de1  # Makes imports "work" for stand-alone use
from pyDE1.bledev.managed_bleak_device import ManagedBleakDevice
from pyDE1.config import config
from pyDE1.de1.c_api import API_MachineStates
from pyDE1.event_manager.event_manager import SubscribedEvent
from pyDE1.event_manager.events import DeviceRole
from pyDE1.event_manager.payloads import EventPayload
from pyDE1.lock_logger import LockLogger
from pyDE1.scanner import find_first_matching, RegisteredPrefixes


TIMEOUT_FIRST_UPDATE = 2.5 # seconds after connect


RegisteredPrefixes.add_to_role('BlueDOT', DeviceRole.THERMOMETER)


class BlueDOTUpdate (EventPayload):

    def __init__(self,
                 arrival_time: float,
                 create_time: Optional[float] = None,
                 ):
        super().__init__(arrival_time=arrival_time,
                         create_time=create_time)
        self._version = "1.0.0"
        self.temperature: Optional[float] = None
        self.high_alarm: Optional[float] = None
        self.units: str = "C"
        self.alarm_byte: Optional[Union[bytearray, int]] = None
        self.name: Optional[str] = None


class BDNotification (NamedTuple):
    arrival_time: float
    raw_data: bytearray
    temperature: float
    high_alarm: float
    units: "BDUnit"
    alarm_byte: Union[bytearray, int]


class BDUnit (enum.IntFlag):
    C = 0
    F = 1

    @property
    def freezing(self):
        if self == BDUnit.C:
            return 0
        else:
            return 32


class BlueDOT (ManagedBleakDevice):

    _de1 = pyDE1.de1.DE1()  # To determine if in a steam cycle when connecting

    def __init__(self):

        self._role = DeviceRole.THERMOMETER
        self.logger = pyDE1.getLogger('BlueDOT')
        super().__init__()

        self.updates = SubscribedEvent(self)

        self._last_update: Optional[BDNotification] = None

        self._have_high_alarm = asyncio.Event()  # For beep-on-connect
        
        # CUUID locks
        self._interval_lock = asyncio.Lock()
        self._updates_lock = asyncio.Lock()
        self._high_alarm_lock = asyncio.Lock()
        self._low_alarm_lock = asyncio.Lock()
        self._units_lock = asyncio.Lock()


    @property
    def temperature(self):
        return self._last_update and self._last_update.temperature

    @property
    def units(self):
        return self._last_update and self._last_update.units

    @property
    def high_alarm(self):
        return self._last_update and self._last_update.high_alarm

    @property
    def last_update(self):
        return self._last_update
        
    async def set_updates_on(self, state=True):
        if not self.is_connected:
            self.logger.warning(
                f"Not connected, can't set_updates_on({state})")
            return
        ll = LockLogger(self._updates_lock, 'updates_lock').check()
        async with self._updates_lock:
            ll.acquired()
            if state:
                await self._bleak_client.start_notify(
                    CUUID.UPDATES.uuid,
                    self._notification_callback)
            else:
                await self._bleak_client.stop_notify(CUUID.UPDATES.uuid)
            self.logger.info(f"Notification enabled: {state}")
        ll.released()
        
    async def set_update_rate(self, period: Union[int, float]):
        if not self._bleak_client.is_connected:
            self.logger.warning(
                f"Not connected, can't set_update_rate({period})")
            return
        ll = LockLogger(self._interval_lock, 'interval_lock').check()
        async with self._interval_lock:
            ll.acquired()
            val = int(round(period))
            await self._bleak_client.write_gatt_char(
                CUUID.INTERVAL.uuid, pack('B', val))
            self.logger.info(f"Notification period set to {val}")
        ll.released()

    async def set_high_alarm(self, temperature: Union[int, float]):
        if not self._bleak_client.is_connected:
            self.logger.warning(
                f"Not connected, can't set_high_alarm({temperature})")
            return
        ll = LockLogger(self._high_alarm_lock, 'high_alarm_lock').check()
        async with self._high_alarm_lock:
            ll.acquired()
            val = int(round(temperature))
            try:
                await self._bleak_client.write_gatt_char(
                    CUUID.HIGH_ALARM.uuid, pack('B', val))
                self.logger.info(f"High alarm set: {val}")
            except struct.error as e:
                self.logger.exception(
                    f"Unable to set high alarm for {temperature}",
                    exc_info=e)
        ll.released()

    async def sample_slow(self):
        await self.set_update_rate(60)

    async def sample_normal(self):
        await self.set_update_rate(config.steam.IDLE_SECONDS_PER_SAMPLE)

    async def sample_fast(self):
        await self.set_update_rate(1)

    async def release(self, timeout: Optional[float] = None) -> bool:
        # super().release() doesn't call this class' request_release()
        if self.is_ready:
            await self.sample_slow()
        return await super().release(timeout)

    async def request_release(self):
        if self.is_ready:
            await self.sample_slow()
        await super().request_release()

    # For external API

    async def change_to_id(self, ble_device_id: Optional[str]):
        """
        For now, this won't return until connected or fails to connect
        As a result, will trigger the timeout on API calls
        """
        self.logger.info(f"Request to replace thermometer with {ble_device_id}")

        if ble_device_id == 'scan':
            await self.connect_to_first_if_found()
        else:
            await self.change_address(ble_device_id)

        return self.address

    async def connect_to_first_if_found(self):
        if self.is_connected:
            self.logger.warning(
                "'scan' requested, but already connected. "
                "No action taken.")
        else:
            device = await find_first_matching(DeviceRole.THERMOMETER)
            if device:
                await self.change_address(device)
                await self.capture()
        return self.address

    # From managed_bleak_device, used by external API
    #   address
    #   name
    #   availability_state
    #   availability_setter
    #   device_availability_last_sent

    ###
    ### End of public API
    ###

    # TODO: "make sure" that go_slow() gets called on exit

    async def _initialize_after_connection(self, hold_ready=False):
        self._have_high_alarm.clear()
        self._last_update = None
        await self.set_updates_on()
        # await asyncio.sleep(1)  # Is this needed, and, if so, when?
        await self.sample_fast()
        try:
            await asyncio.wait_for(
                self._have_high_alarm.wait(),
                timeout=TIMEOUT_FIRST_UPDATE)
            # Beep the alarm on getting the update
            old_ha = self._last_update.high_alarm
            new_ha = self._last_update.units.freezing
            await self.set_high_alarm(new_ha)
            await asyncio.sleep(1.0)  # beep time
            await self.set_high_alarm(old_ha)
        except asyncio.TimeoutError:
            self.logger.warning(
                f"Did not get update within {TIMEOUT_FIRST_UPDATE:.1f} sec, "
                "no beep for connection.")
            pass
        if (self._de1.is_ready
                and self._de1.current_state == API_MachineStates.Steam):
            await self.sample_fast()
        else:
            await self.sample_normal()
        self._notify_ready()

    async def _notification_callback(self, sender: int, data: bytearray):

        now = time.time()

        byte_0 = data[0:1].hex()
        (current, high_alarm) = unpack_from('<ii', data, offset=1)
        (units_byte,) = unpack_from('B', data, offset=11)
        byte_12 = data[12:13].hex()
        bluedot_mac = data[13:19]
        alarm_byte = data[19:20]

        # Only used for logging
        last_time = self._last_update.arrival_time if self._last_update else None

        self._last_update = BDNotification(
            arrival_time=now,
            raw_data=data,
            temperature=current,
            high_alarm=high_alarm,
            units=BDUnit(units_byte),
            alarm_byte=alarm_byte,
        )
        self._have_high_alarm.set()

        payload = BlueDOTUpdate(arrival_time=self._last_update.arrival_time)
        payload.temperature = self._last_update.temperature
        payload.high_alarm = self._last_update.high_alarm
        payload.units = self._last_update.units.name
        payload.alarm_byte = self._last_update.alarm_byte
        payload.name = self.name
        await self.updates.publish(payload)

        units_string = BDUnit(units_byte).name
        dt = now - last_time if last_time else 0

        self.logger.debug(
            f"{dt:0.3f} {current} {high_alarm} {units_string} "
            f"{byte_0} {byte_12} {alarm_byte.hex()}")

        

class CUUID (enum.Enum):
    INTERVAL    = '60721D99-6698-4EEC-8E0A-50D0C37F17B9'
    UPDATES     = '783F2991-23E0-4BDC-AC16-78601BD84B39'
    HIGH_ALARM  = 'DE0415CF-D54A-4EA4-A58F-C7AA07F79BAA'
    LOW_ALARM   = 'BFDBEB45-11A3-4406-BCB4-ED7C6F939FBC'
    UNITS       = 'C86CF012-5C33-48AA-82D1-84A57C908CA0'
        
    @property 
    def uuid(self):
        return self.value


if __name__ == "__main__":

    import argparse
    from pyDE1 import pyde1_logging
    from pyDE1.config import config
    import pyDE1.shutdown_manager as sm

    pyde1_logging.setup_initial_logger()

    ap = argparse.ArgumentParser(
        description="""Test the BLueDOT alone.

        """
        f"Default configuration file is at {pyDE1.config.DEFAULT_CONFIG_FILE}"
    )
    ap.add_argument('-c', type=str, help='Use as alternate config file')

    args = ap.parse_args()

    config.load_from_yaml(args.c)

    config.logging.handlers.STDERR = 'DEBUG'
    config.logging.formatters.STDERR = config.logging.formatters.LOGFILE

    pyde1_logging.setup_direct_logging(config.logging)
    pyde1_logging.config_logger_levels(config.logging)

    logger = pyDE1.getLogger('Main')


    loop = asyncio.get_event_loop()
    loop.set_debug(True)
    loop.set_exception_handler(sm.exception_handler)
    sm.attach_signal_handler_to_loop(sm.shutdown, loop)

    async def setup_and_run():
        bluetooth_id = '00:a0:50:e2:f6:49'
        bd = BlueDOT()

        # async def wait_for_shutdown():
        #     await sm.wait_for_shutdown_underway()
        #     print("shutdown")
        #     await bd.go_slow()
        #     print("go slow done")
        #     await asyncio.sleep(0.5)
        #     sm.cleanup_complete.set()
        # loop.create_task(wait_for_shutdown())

        await bd.change_address(bluetooth_id)
        await bd.request_capture()
        await bd.event_ready.wait()
        await bd.sample_fast()
        await asyncio.sleep(600)
        await bd.release()
        await asyncio.sleep(5)

    loop.run_until_complete(setup_and_run())